"""This module describes which functions are intended to be used by users to
create papis scripts.
"""

import logging

logger = logging.getLogger("api")
logger.debug("importing")

import os
import papis.utils
import papis.commands
import papis.config
import papis.database


class status():
    success = 0
    generic_fail = 1
    file_not_found = 2


def get_lib():
    """Get current library, it either retrieves the library from
    the environment PAPIS_LIB variable or from the command line
    args passed by the user.

    :returns: Library name
    :rtype:  str

    >>> get_lib() is not None
    True
    """
    return papis.config.get_lib()


def set_lib(library):
    """Set current library, it either sets the library in
    the environment PAPIS_LIB variable or in the command line
    args passed by the user.

    :param library: Name of library or path to a given library
    :type  library: str

    """
    return papis.config.set_lib(library)


def get_arg(arg, default=None):
    try:
        val = getattr(papis.commands.get_args(), arg)
    except AttributeError:
        try:
            val = os.environ["PAPIS_"+arg.upper()]
        except KeyError:
            val = default
    return val


def get_libraries():
    """Get all libraries declared in the configuration. A library is discovered
    if the ``dir`` key defined in the library section.

    :returns: List of library names
    :rtype: list

    >>> len(get_libraries()) >= 1
    True

    """
    libs = []
    config = papis.config.get_configuration()
    for key in config.keys():
        if "dir" in config[key]:
            libs.append(key)
    return libs


def pick_doc(documents: list):
    """Pick a document from documents with the correct formatting

    :documents: List of documents
    :returns: Document

    >>> from papis.document import from_data
    >>> doc = from_data({'title': 'Hello World'})
    >>> pick_doc([doc]).dump()
    'title:   Hello World\\n'

    """
    header_format_path = papis.config.get('header-format-file')
    if header_format_path is not None:
        with open(os.path.expanduser(header_format_path)) as fd:
            header_format = fd.read()
    else:
        header_format = papis.config.get("header-format")
    match_format = papis.config.get("match-format")
    pick_config = dict(
        header_filter=lambda x: papis.utils.format_doc(header_format, x),
        match_filter=lambda x: papis.utils.format_doc(match_format, x)
    )
    return papis.api.pick(
        documents,
        pick_config
    )


def pick(options: list, pick_config={}):
    """This is a wrapper for the various pickers that are supported.
    Depending on the configuration different selectors or 'pickers'
    are used.

    :param options: List of different objects. The type of the objects within
        the list must be supported by the pickers. This is the reason why this
        function is difficult to generalize for external picker programs.
    :type  options: list

    :param pick_config: Dictionary with additional configuration for the used
        picker. This depends on the picker.
    :type  pick_config: dict

    :returns: Returns elements of ``options``.
    :rtype: Element(s) of ``options``

    >>> papis.config.set('picktool', 'papis.pick')
    >>> pick(['something'])
    'something'
    >>> papis.config.set('picktool', 'nonexistent')
    >>> pick(['something'])
    Traceback (most recent call last):
    ...
    Exception: I don\'t know how to use the picker \'nonexistent\'
    >>> papis.config.set('picktool', 'papis.pick')

    """
    # Leave this import here
    import papis.config
    logger.debug("Parsing picktool")
    external_picker = papis.config.get_external_picker()
    picker = external_picker or papis.config.get("picktool")
    if picker == "papis.pick":
        import papis.pick
        logger.debug("Using papis.pick picker")
        return papis.pick.pick(options, **pick_config)
    elif papis.config.get_external_picker() is not None:
        return papis.config.get_external_picker()(options, **pick_config)
    else:
        raise Exception("I don't know how to use the picker '%s'" % picker)


def open_file(file_path, wait=True):
    """Open file using the ``opentool`` key value as a program to
    handle file_path.

    :param file_path: File path to be handled.
    :type  file_path: str
    :param wait: Wait for the completion of the opener program to continue
    :type  wait: bool

    """
    papis.utils.general_open(file_path, "opentool", wait=wait)


def open_dir(dir_path, wait=True):
    """Open dir using the ``file-browser`` key value as a program to
    open dir_path.

    :param dir_path: Folder path to be handled.
    :type  dir_path: str
    :param wait: Wait for the completion of the opener program to continue
    :type  wait: bool

    """
    papis.utils.general_open(dir_path, "file-browser", wait=wait)


def edit_file(file_path, wait=True):
    """Edit file using the ``editor`` key value as a program to
    handle file_path.

    :param file_path: File path to be handled.
    :type  file_path: str
    :param wait: Wait for the completion of the opener program to continue
    :type  wait: bool

    """
    papis.utils.general_open(file_path, "editor", wait=wait)


def get_all_documents_in_lib(library=None):
    """Get ALL documents contained in the given library with possibly.

    :param library: Library name.
    :type  library: str

    :returns: List of all documents.
    :rtype: list

    >>> import tempfile
    >>> folder = tempfile.mkdtemp()
    >>> set_lib(folder)
    >>> docs = get_all_documents_in_lib(folder)
    >>> len(docs)
    0

    """
    return papis.database.get(library=library).get_all_documents()


def get_documents_in_dir(directory, search=""):
    """Get documents contained in the given folder with possibly a search
    string.

    :param directory: Folder path.
    :type  directory: str

    :param search: Search string
    :type  search: str

    :returns: List of filtered documents.
    :rtype: list

    >>> import tempfile
    >>> docs = get_documents_in_dir(tempfile.mkdtemp())
    >>> len(docs)
    0

    """
    set_lib(directory)
    return get_documents_in_lib(directory, search)


def get_documents_in_lib(library=None, search=""):
    """Get documents contained in the given library with possibly a search
    string.

    :param library: Library name.
    :type  library: str

    :param search: Search string
    :type  search: str

    :returns: List of filtered documents.
    :rtype: list

    """
    return papis.database.get(library=library).query(search)


def clear_lib_cache(lib=None):
    """Clear cache associated with a library. If no library is given
    then the current library is used.

    :param lib: Library name.
    :type  lib: str

    >>> clear_lib_cache()

    """
    lib = papis.api.get_lib() if lib is None else lib
    papis.database.get(lib).clear()



"""
TODO: Create an own python package for this

For description refer to
https://www.base-search.net/about/download/base_interface.pdf

"""
import urllib.parse
import urllib.request  # import urlencode
import papis.config
import logging
import json

logger = logging.getLogger('base')

BASE_BASEURL = "https://api.base-search.net/cgi-bin/BaseHttpSearchInterface.fcgi/"

def get_data(query="", max=20):

    logger.warning('BASE engine in papis is experimental')

    if max > 125:
        logger.error('BASE only allows a maximum of 125 hits')
        max = 125

    results = []
    dict_params = {
        "func": "PerformSearch",
        "query": query,
        "format": "json",
        "hits": max,
    }
    params = urllib.parse.urlencode(
        {x: dict_params[x] for x in dict_params if dict_params[x]}
    )
    req_url = BASE_BASEURL + "search?" + params
    logger.debug("url = " + req_url)
    url = urllib.request.Request(
        req_url,
        headers={
            'User-Agent': 'papis'
        }
    )
    jsondoc = json.loads(urllib.request.urlopen(url).read().decode())
    docs = jsondoc.get('response').get('docs')
    logger.info("Retrieved {0} documents".format(len(docs)))
    return list(map(basedoc_to_papisdoc, docs))


def basedoc_to_papisdoc(basedoc):
    """Convert a json doc from the base database into a papis document

    :basedoc: Json doc from base database
    :returns: Dictionary containing its data

    """
    doc = dict()
    keys_translate = [
        ("dctitle", "title", "single"),
        ("dcyear", "year", "single"),
        ("dclink", "url", "single"),
        ("dcdescription", "abstract", "single"),
        ("dcpublisher", "publisher", "multi", lambda x: x[0]),
        ("dcperson", "author", "multi", lambda x: " and ".join(x)),
        ("dcsubject", "tags", "multi", lambda x: " ".join(x)),
        ("dcdoi", "doi", "multi", lambda x: x[0]),
        ("dctype", "type", "multi", lambda x: x[0].lower()),
        ("dclang", "lang", "mutli", lambda x: x[0]),
    ]
    for kt in keys_translate:
        if kt[0] in basedoc.keys():
            key = kt[1]
            if kt[2] == "multi":
                value = basedoc[kt[0]]
                value = kt[3](value) if kt[3] is not None else value
            else:
                value = basedoc[kt[0]]
            doc[key] = value
    return doc



import re
import papis.downloaders.base
import bs4


class Downloader(papis.downloaders.base.Downloader):

    def __init__(self, url):
        papis.downloaders.base.Downloader.__init__(self, url, name="hal")
        self.expected_document_extension = 'pdf'

    @classmethod
    def match(cls, url):
        if re.match(r".*hal.*\.fr.*", url):
            return Downloader(url)
        else:
            return False

    def get_identifier(self):
        """
        >>> d = Downloader("http://www.hal.fr/2014TOU30305")
        >>> d.get_identifier()
        '2014TOU30305'
        >>> d = Downloader("http://www.hal.fr/2014TOU30305.bib/?asdf=2")
        >>> d.get_identifier()
        '2014TOU30305'
        """
        m = re.match(r".*hal.fr/([^/?.&]+).*", self.url)
        return m.group(1) if m is not None else None

    def get_document_url(self):
        """
        >>> d = Downloader("https://hal.archives-ouvertes.fr/jpa-00205888?asf=")
        >>> d.get_document_url()
        'https://hal.archives-ouvertes.fr/jpa-00205888/document'
        """
        url = re.sub(r'\?.*', '', self.get_url()) + '/document'
        return url

    def get_bibtex_url(self):
        """
        >>> d = Downloader("https://hal.archives-ouvertes.fr/jpa-00205888?asf=")
        >>> d.get_bibtex_url()
        'https://hal.archives-ouvertes.fr/jpa-00205888/bibtex'
        """
        url = re.sub(r'\?.*', '', self.get_url()) + '/bibtex'
        return url



import re
from prompt_toolkit.application.current import get_app
from prompt_toolkit.formatted_text.html import HTML, html_escape
from prompt_toolkit.filters import Condition
from prompt_toolkit.layout.screen import Point
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.layout.controls import (
    FormattedTextControl, BufferControl
)
from prompt_toolkit.widgets import HorizontalLine
from prompt_toolkit.layout.containers import (
    HSplit, Window, ConditionalContainer, WindowAlign, ScrollOffsets
)
from prompt_toolkit.filters import has_focus
from prompt_toolkit.layout import NumberedMargin
import multiprocessing

import logging

logger = logging.getLogger('tui:widget:list')


class OptionsList(ConditionalContainer):

    def __init__(
            self,
            options,
            default_index=0,
            header_filter=lambda x: x,
            match_filter=lambda x: x,
            custom_filter=None,
            search_buffer=Buffer(multiline=False)
            ):

        assert(isinstance(options, list))
        assert(callable(header_filter))
        assert(callable(match_filter))
        assert(isinstance(default_index, int))

        self.search_buffer = search_buffer
        self.search_buffer.on_text_changed += self.update

        self.header_filter = header_filter
        self.match_filter = match_filter
        self.current_index = default_index
        self.entries_left_offset = 0
        self.pool = multiprocessing.Pool(multiprocessing.cpu_count())

        self._options = []
        self.max_entry_height = 1
        # Options are processed here also through the setter
        self.options = options
        self.cursor = Point(0, 0)

        self.content = FormattedTextControl(
            text=self.get_tokens,
            focusable=False,
            key_bindings=None,
            get_cursor_position=lambda: self.cursor,
        )
        self.content_window = Window(
            content=self.content,
            wrap_lines=False,
            allow_scroll_beyond_bottom=True,
            scroll_offsets=ScrollOffsets(bottom=self.max_entry_height),
            cursorline=False,
            cursorcolumn=False,
            #right_margins=[NumberedMargin()],
            #left_margins=[NumberedMargin()],
            align=WindowAlign.LEFT,
            height=None,
            get_line_prefix=self.get_line_prefix
            #get_line_prefix=lambda line, b: [('bg:red', '  ')]
        )

        self.update()

        super(OptionsList, self).__init__(
            content=self.content_window,
            filter=(
                custom_filter
                if custom_filter is not None
                else has_focus(self.search_buffer)
            )
        )

    def __del__(self):
        # Clean initialized pool upon deleting of the object
        self.pool.close()
        self.pool.join()

    def get_line_prefix(self, line, blih):
        try:
            index = self.indices.index(self.current_index)
            cline = sum(
                self.options_headers_linecount[i]
                for i in self.indices[0:index]
            )
            cline = sum(
                self.options_headers_linecount[i]
                for i in self.indices[0:index]
            )
            if line == cline:
                return [('class:options_list.selected_margin', '>')]
            else:
                return [('class:options_list.unselected_margin', ' ')]
        except:
            return

    @property
    def options(self):
        return self._options

    @options.setter
    def options(self, new_options):
        self._options = new_options
        self.process_options()

    def move_up(self):
        try:
            index = self.indices.index(self.current_index)
            index -= 1
            if index < 0:
                self.current_index = self.indices[-1]
            else:
                self.current_index = self.indices[index]
        except ValueError:
            pass

    def move_down(self):
        try:
            index = self.indices.index(self.current_index)
            index += 1
            if index >= len(self.indices):
                self.current_index = self.indices[0]
            else:
                self.current_index = self.indices[index]
        except ValueError:
            pass

    def go_top(self):
        if len(self.indices) > 0:
            self.current_index = self.indices[0]

    def go_bottom(self):
        if len(self.indices) > 0:
            self.current_index = self.indices[-1]

    @property
    def search_regex(self):
        cleaned_search = (
            self.search_buffer.text
            .replace('(', '\\(')
            .replace(')', '\\)')
            .replace('+', '\\+')
            .replace('[', '\\[')
            .replace(']', '\\]')
        )
        return re.compile(r".*"+re.sub(r"\s+", ".*", cleaned_search), re.I)

    @staticmethod
    def match_against_regex(regex, line, index):
        return index if regex.match(line, re.I) else None

    def update(self, *args):
        self.filter_options()

    def filter_options(self, *args):
        indices = []
        regex = self.search_regex

        result = [
            self.pool.apply_async(
                OptionsList.match_against_regex,
                args=(regex, opt, i,)
            ) for i, opt in enumerate(self.options_matchers)
        ]

        indices =  [d.get() for d in result if d.get() is not None]

        self.indices = indices
        if len(self.indices) and self.current_index not in self.indices:
            if self.current_index < min(self.indices):
                self.current_index = self.indices[0]
            elif self.current_index > max(self.indices):
                self.current_index = max(self.indices)
            else:
                self.current_index = self.indices[0]

    def get_selection(self):
        if len(self.indices) and self.current_index is not None:
            return self.options[self.current_index]

    def update_cursor(self):
        """This function updates the cursor according to the current index
        in the list.
        """
        try:
            index = self.indices.index(self.current_index)
            line = sum(
                self.options_headers_linecount[i]
                for i in self.indices[0:index]
            )
            self.cursor = Point(0, line)
        except Exception as e:
            self.cursor = Point(0, 0)

    def get_tokens(self):
        self.update_cursor()
        result = sum(
            [self.options_headers[i] for i in self.indices],
            []
        )
        return result

    def process_options(self):
        logger.debug('processing {0} options'.format(len(self.options)))
        self.options_headers_linecount = [
            len(self.header_filter(o).split('\n'))
            for o in self.options
        ]
        self.max_entry_height = max(self.options_headers_linecount)
        logger.debug('processing headers')
        self.options_headers = []
        for o in self.options:
            prestring = self.header_filter(o) + '\n'
            try:
                htmlobject = HTML(prestring).formatted_text
            except:
                logger.error(
                    'Error processing html for \n {0}'.format(prestring)
                )
                htmlobject = [ ('fg:red', prestring) ]
            self.options_headers += [htmlobject]
        logger.debug('got {0} headers'.format(len(self.options_headers)))
        logger.debug('processing matchers')
        self.options_matchers = [self.match_filter(o) for o in self.options]
        self.indices = range(len(self.options))
        logger.debug('got {0} matchers'.format(len(self.options_matchers)))



import os
import re
from prompt_toolkit.application import Application
from prompt_toolkit.formatted_text.html import HTML
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.layout.processors import BeforeInput
from prompt_toolkit.key_binding import KeyBindings, merge_key_bindings
from prompt_toolkit.layout.screen import Point
from prompt_toolkit.filters import has_focus, Condition
from prompt_toolkit.styles import Style
from prompt_toolkit.layout.containers import (
    HSplit, Window, ConditionalContainer
)
from prompt_toolkit.layout.controls import (
    BufferControl,
)
from prompt_toolkit.layout.layout import Layout
import papis.config as config
import logging

from .widgets.command_line_prompt import Command
from .widgets import (
    InfoWindow, CommandLinePrompt, HelpWindow, OptionsList,
    MessageToolbar
)

logger = logging.getLogger('pick')

_keys_info = None
def get_keys_info():
    global _keys_info
    if not _keys_info:
        _keys_info = {
            "move_down_key": {
                'key': config.get('move_down_key', section='tui'),
                'help': 'Move cursor down in the list',
            },
            "move_up_key": {
                'key': config.get('move_up_key', section='tui'),
                'help': 'Move cursor up in the list',
            },
            "move_down_while_info_window_active_key": {
                'key': config.get('move_down_while_info_window_active_key', section='tui'),
                'help': 'Move cursor down while info window is active',
            },
            "move_up_while_info_window_active_key": {
                'key': config.get('move_up_while_info_window_active_key', section='tui'),
                'help': 'Move cursor up while info window is active',
            },
            "focus_command_line_key": {
                'key': config.get('focus_command_line_key', section='tui'),
                'help': 'Focus command line prompt',
            },
            "edit_document_key": {
                'key': config.get('edit_document_key', section='tui'),
                'help': 'Edit currently selected document',
            },
            "open_document_key": {
                'key': config.get('open_document_key', section='tui'),
                'help': 'Open currently selected document',
            },
            "show_help_key": {
                'key': config.get('show_help_key', section='tui'),
                'help': 'Show help',
            },
            "show_info_key": {
                'key': config.get('show_info_key', section='tui'),
                'help': 'Show the yaml information of the current document',
            },
            "go_top_key": {
                'key': config.get('go_top_key', section='tui'),
                'help': 'Go to the top of the list',
            },
            "go_bottom_key": {
                'key': config.get('go_bottom_key', section='tui'),
                'help': 'Go to the bottom of the list',
            },
        }
    return _keys_info


def create_keybindings(app):
    keys_info = get_keys_info()
    kb = KeyBindings()

    @kb.add('escape', filter=Condition(lambda: app.message_toolbar.text))
    def _(event):
        event.app.message_toolbar.text = None

    @kb.add('escape', filter=Condition(lambda: app.error_toolbar.text))
    def _(event):
        event.app.error_toolbar.text = None

    @kb.add('c-n', filter=~has_focus(app.info_window))
    @kb.add(keys_info["move_down_key"]["key"], filter=~has_focus(app.info_window))
    def down_(event):
        event.app.options_list.move_down()
        event.app.refresh()
        event.app.update()

    @kb.add(
        keys_info["move_down_while_info_window_active_key"]["key"],
        filter=has_focus(app.info_window)
    )
    def down_info(event):
        down_(event)
        event.app.update_info_window()

    @kb.add('c-p', filter=~has_focus(app.info_window))
    @kb.add(keys_info["move_up_key"]["key"], filter=~has_focus(app.info_window))
    def up_(event):
        event.app.options_list.move_up()
        event.app.refresh()
        event.app.update()

    @kb.add(
        keys_info["move_up_while_info_window_active_key"]["key"],
        filter=has_focus(app.info_window)
    )
    def up_info(event):
        up_(event)
        event.app.update_info_window()

    @kb.add('q', filter=has_focus(app.help_window))
    @kb.add('escape', filter=has_focus(app.help_window))
    def _help_quit(event):
        event.app.layout.focus(app.help_window.window)
        event.app.layout.focus(app.command_line_prompt.window)
        event.app.message_toolbar.text = None
        event.app.layout.focus(event.app.options_list.search_buffer)

    @kb.add('q', filter=has_focus(app.info_window))
    @kb.add('s-tab', filter=has_focus(app.info_window))
    @kb.add('escape', filter=has_focus(app.info_window))
    def _info(event):
        event.app.layout.focus(event.app.options_list.search_buffer)
        event.app.message_toolbar.text = None

    @kb.add(
        keys_info["focus_command_line_key"]["key"],
        filter=~has_focus(app.command_line_prompt)
    )
    def _command_window(event):
        event.app.layout.focus(app.command_line_prompt.window)

    @kb.add('enter', filter=has_focus(app.command_line_prompt))
    def _enter_(event):
        event.app.layout.focus(event.app.options_list.search_buffer)
        try:
            event.app.command_line_prompt.trigger()
        except Exception as e:
            event.app.error_toolbar.text = str(e)
        event.app.command_line_prompt.clear()

    @kb.add('escape', filter=has_focus(app.command_line_prompt))
    def _(event):
        event.app.layout.focus(event.app.options_list.search_buffer)
        event.app.command_line_prompt.clear()

    return kb


def get_commands(app):

    kb = KeyBindings()
    keys_info = get_keys_info()

    @kb.add('c-q')
    @kb.add('c-c')
    def exit(event):
        event.app.deselect()
        event.app.exit()

    @kb.add(
        'enter',
        filter=has_focus(app.options_list.search_buffer)
    )
    def select(event):
        event.app.exit()

    @kb.add(
        keys_info["open_document_key"]["key"],
        filter=has_focus(app.options_list.search_buffer)
    )
    def open(cmd):
        from papis.commands.open import run
        doc = cmd.app.get_selection()
        run(doc)

    @kb.add(
        keys_info["edit_document_key"]["key"],
        filter=has_focus(app.options_list.search_buffer)
    )
    def edit(cmd):
        from papis.commands.edit import run
        doc = cmd.app.get_selection()
        run(doc)
        cmd.app.renderer.clear()

    @kb.add(
        keys_info["show_help_key"]["key"],
        filter=~has_focus(app.help_window)
    )
    def help(event):
        event.app.layout.focus(app.help_window.window)
        event.app.message_toolbar.text = 'Press q to quit'

    def _echo(cmd, *args):
        cmd.app.message_toolbar.text = ' '.join(args)

    @kb.add(
        keys_info["show_info_key"]["key"],
        filter=~has_focus(app.info_window)
    )
    def info(cmd):
        cmd.app.update_info_window()
        cmd.app.layout.focus(cmd.app.info_window.window)

    @kb.add('c-g', 'g')
    @kb.add(keys_info["go_top_key"]["key"])
    def go_top(event):
        event.app.options_list.go_top()
        event.app.refresh()

    @kb.add('c-g', 'G')
    @kb.add(keys_info["go_bottom_key"]["key"])
    def go_end(event):
        event.app.options_list.go_bottom()
        event.app.refresh()

    return ([
        Command("open", run=open, aliases=["op"]),
        Command("edit", run=edit, aliases=["e"]),
        Command("select", run=select, aliases=["e"]),
        Command("exit", run=exit, aliases=["quit", "q"]),
        Command("info", run=info, aliases=["i"]),
        Command("go_top", run=go_top),
        Command("go_bottom", run=go_end),
        Command("move_down", run=lambda c: c.app.options_list.move_down()),
        Command("move_up", run=lambda c: c.app.options_list.move_up()),
        Command("echo", run=_echo),
        Command("help", run=help),
    ], kb)


class Picker(Application):
    """The :class:`Picker <Picker>` object

    :param options: a list of options to choose from
    :param default_index: (optional) set this if the default
        selected option is not the first one
    """

    def __init__(
            self,
            options,
            default_index=0,
            header_filter=lambda x: x,
            match_filter=lambda x: x
            ):


        self.info_window = InfoWindow()
        self.help_window = HelpWindow()
        self.message_toolbar = MessageToolbar(style="class:message_toolbar")
        self.error_toolbar = MessageToolbar(style="class:error_toolbar")
        self.status_line = MessageToolbar(style="class:status_line")
        self.status_line_format = config.get(
            'status_line_format', section="tui"
        )

        self.options_list = OptionsList(
            options,
            default_index,
            header_filter,
            match_filter,
            custom_filter=~has_focus(self.help_window)
        )

        commands, commands_kb = get_commands(self)
        self.command_line_prompt = CommandLinePrompt(commands=commands)
        kb = merge_key_bindings([create_keybindings(self), commands_kb])

        _root_container = HSplit([
            HSplit([
                Window(
                    content=BufferControl(
                        input_processors=[BeforeInput('> ')],
                        buffer=self.options_list.search_buffer
                    )
                ),
                self.options_list,
                self.info_window,
            ]),
            self.help_window,
            self.error_toolbar,
            self.message_toolbar,
            self.status_line,
            self.command_line_prompt.window,
        ])

        regex = re.compile(r'.*\.([^ ]+) +at.*')

        help_text = ""
        keys_info = get_keys_info()
        for k in keys_info:
            help_text += (
                "<ansired>{k[key]}</ansired>: {k[help]}\n".format(
                    k=keys_info[k]
                )
            )
        self.help_window.text = HTML(help_text)

        self.layout = Layout(_root_container)

        super(Picker, self).__init__(
            input=None,
            output=None,
            editing_mode=EditingMode.EMACS
            if config.get('editmode', section='tui') == 'emacs'
            else EditingMode.VI,
            layout=self.layout,
            style=Style.from_dict({
                'options_list.selected_margin': config.get(
                    'options_list.selected_margin_style', section='tui'
                ),
                'options_list.unselected_margin': config.get(
                    'options_list.unselected_margin_style', section='tui'
                ),
                'error_toolbar': config.get(
                    'error_toolbar_style', section='tui'
                ),
                'message_toolbar': config.get(
                    'message_toolbar_style', section='tui'
                ),
                'status_line': config.get(
                    'status_line_style', section='tui'
                ),
            }),
            key_bindings=kb,
            include_default_pygments_style=False,
            full_screen=True,
            enable_page_navigation_bindings=True
        )
        self.update()

    def deselect(self):
        self.options_list.current_index = None

    def refresh_status_line(self):
        self.status_line.text = self.status_line_format.format(
            selected_index=int(self.options_list.current_index) + 1,
            number_of_documents=len(self.options_list.options),
        )

    def refresh(self, *args):
        self.refresh_status_line()

    def update(self, *args):
        self.options_list.update()
        self.refresh_status_line()

    def get_selection(self):
        return self.options_list.get_selection()

    def update_info_window(self):
        doc = self.options_list.get_selection()
        self.info_window.set_text(doc.dump())



from papis.tui.app import Picker


def pick(
        options,
        default_index=0,
        header_filter=lambda x: x,
        match_filter=lambda x: x
        ):
    """Construct and start a :class:`Picker <Picker>`.
    """

    if len(options) == 0:
        return ""
    if len(options) == 1:
        return options[0]

    picker = Picker(
        options,
        default_index,
        header_filter,
        match_filter
    )
    picker.run()
    return picker.options_list.get_selection()



from prompt_toolkit.formatted_text.html import HTML
from prompt_toolkit.layout.processors import BeforeInput
from prompt_toolkit.filters import has_focus, Condition
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.layout.containers import (
    HSplit, Window, WindowAlign, ConditionalContainer
)
from prompt_toolkit.layout import Dimension
from prompt_toolkit.layout.controls import (
    BufferControl,
    FormattedTextControl
)
from prompt_toolkit.widgets import (
    HorizontalLine
)
from prompt_toolkit.lexers import PygmentsLexer
from pygments.lexers import find_lexer_class_by_name
import logging

from .list import OptionsList
from .command_line_prompt import CommandLinePrompt

logger = logging.getLogger('pick')


class MessageToolbar(ConditionalContainer):

    def __init__(self, style=""):
        self.message = None
        self.text_control = FormattedTextControl(text="")
        super(MessageToolbar, self).__init__(
            content=Window(
                style=style, content=self.text_control,
                height=1
            ),
            filter=Condition(lambda: self.text)
        )

    @property
    def text(self):
        return self.text_control.text

    @text.setter
    def text(self, value):
        self.text_control.text = value


class InfoWindow(ConditionalContainer):

    def __init__(self, lexer_name='yaml'):
        self.buf = Buffer()
        self.buf.text = ''
        self.lexer = PygmentsLexer(find_lexer_class_by_name(lexer_name))
        self.window = HSplit([
            HorizontalLine(),
            Window(
                content=BufferControl(
                    buffer=self.buf, lexer=self.lexer)
            )
        ], height=Dimension(min=5, max=20, weight=1))
        super(InfoWindow, self).__init__(
            content=self.window,
            filter=has_focus(self)
        )

    def set_text(self, text):
        self.buf.text = text

    def get_text(self):
        return self.buf.text


class HelpWindow(ConditionalContainer):
    help_text = ''
    # help_text = """
    # <span fg='ansired'> Bindings: </span>

# ----------------------------------------------------
# /   Ctrl-e, Ctrl-down, Shift-down : Scroll Down     /
# /   Ctrl-y, Ctrl-up,   Shift-up   : Scroll up       /
# /   Ctrl-n, down                  : Next item       /
# /   Ctrl-p, up                    : Previous item   /
# /   Ctrl-q, Ctrl-c                : Quit            /
# /   Home                          : First item      /
# /   End                           : Last item       /
# ----------------------------------------------------
    # """

    def __init__(self):
        self.text_control = FormattedTextControl(
            text=HTML(self.help_text)
        )
        self.window = Window(
            content=self.text_control,
            always_hide_cursor=True,
            align=WindowAlign.LEFT
        )
        super(HelpWindow, self).__init__(
            content=self.window,
            filter=has_focus(self.window)
        )

    @property
    def text(self):
        return self.text_control.text

    @text.setter
    def text(self, value):
        self.text_control.text = value



from prompt_toolkit.filters import has_focus, Condition
from prompt_toolkit.layout.processors import BeforeInput
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.completion import WordCompleter, Completer, Completion
from prompt_toolkit.layout.containers import (
    HSplit, Window, WindowAlign, ConditionalContainer
)
from prompt_toolkit.layout.controls import (
    BufferControl,
    FormattedTextControl
)
from prompt_toolkit.application.current import get_app
from prompt_toolkit.utils import Event
import shlex


class Command(Event):
    """
    :param name: Name of the command
    :type  name: parameter_type
    :param run: A callable object where the first argument is the cmd itself.
    :type  run: callable
    """
    def __init__(self, name, run, aliases=[]):
        assert(isinstance(name, str)), 'name should be a string'
        assert(callable(run)), 'run should be callable'
        assert(isinstance(aliases, list))
        self.name = name
        self.run = run
        self.aliases = aliases

    @property
    def app(self):
        return get_app()

    @property
    def names(self):
        return [self.name] + self.aliases

    def __call__(self, *args, **kwargs):
        return self.run(self, *args, **kwargs)


class CommandLinePrompt(ConditionalContainer):
    """
    A vim-like command line prompt widget.
    It's supposed to be instantiated only once.
    """
    def __init__(self, commands=[]):
        assert(isinstance(commands, list))
        for c in commands:
            assert(isinstance(c, Command))
        self.commands = commands
        wc = WordCompleter(sum([c.names for c in commands], []))
        self.buf = Buffer(
            completer=wc, complete_while_typing=True
        )
        self.buf.text = ''
        self.window = Window(
            content=BufferControl(
                buffer=self.buf,
                input_processors=[BeforeInput(':')]
            ),
            height=1
        )
        super(CommandLinePrompt, self).__init__(
            content=self.window,
            filter=has_focus(self.window)
        )

    def trigger(self):
        input_cmd = shlex.split(self.buf.text)
        if not input_cmd:
            return
        name = input_cmd[0]
        cmds = list(filter(lambda c: name in c.names, self.commands))

        if len(cmds) > 1:
            raise Exception('More than one command matches the input')
        elif len(cmds) == 0:
            raise Exception('No command found ({0})'.format(name))

        input_cmd.pop(0)
        return cmds[0](*input_cmd)

    def clear(self):
        self.text = ''

    @property
    def text(self):
        return self.buf.text

    @text.setter
    def text(self, text):
        self.buf.text = text




def get_default_settings():
    return dict(tui={
        "status_line_format": (
            "{selected_index}/{number_of_documents}  " +
            "F1:help  " +
            "c-l:redraw  "
        ),

        "status_line_style": 'bg:ansiwhite fg:ansiblack',
        'message_toolbar_style': 'bg:ansiyellow fg:ansiblack',
        'options_list.selected_margin_style': 'bg:ansiblack fg:ansigreen',
        'options_list.unselected_margin_style': 'bg:ansiwhite',
        'error_toolbar_style': 'bg:ansired fg:ansiblack',

        'move_down_key': 'down',
        'move_up_key': 'up',
        'move_down_while_info_window_active_key': 'c-n',
        'move_up_while_info_window_active_key': 'c-p',
        'focus_command_line_key': 'tab',
        'edit_document_key': 'c-e',
        'open_document_key': 'c-o',
        'show_help_key': 'f1',
        'show_info_key': 's-tab',
        'go_top_key': 'home',
        'go_bottom_key': 'end',

        "editmode": "emacs",
    })
